@@ -10,10 +10,12 @@ related to writing and reading connections files.
 
 from __future__ import absolute_import
 
+import errno
 import glob
 import json
 import os
 import socket
+import stat
 import tempfile
 import warnings
 from getpass import getpass
@@ -135,6 +137,32 @@ def write_connection_file(fname=None, shell_port=0, iopub_port=0, stdin_port=0,
     with open(fname, 'w') as f:
         f.write(json.dumps(cfg, indent=2))
 
+    if hasattr(stat, 'S_ISVTX'):
+        # set the sticky bit on the file and its parent directory
+        # to avoid periodic cleanup
+        paths = [fname]
+        runtime_dir = os.path.dirname(fname)
+        if runtime_dir:
+            paths.append(runtime_dir)
+        for path in paths:
+            permissions = os.stat(path).st_mode
+            new_permissions = permissions | stat.S_ISVTX
+            if new_permissions != permissions:
+                try:
+                    os.chmod(path, new_permissions)
+                except OSError as e:
+                    if e.errno == errno.EPERM and path == runtime_dir:
+                        # suppress permission errors setting sticky bit on runtime_dir,
+                        # which we may not own.
+                        pass
+                    else:
+                        # failed to set sticky bit, probably not a big deal
+                        warnings.warn(
+                            "Failed to set sticky bit on %r: %s"
+                            "\nProbably not a big deal, but runtime files may be cleaned up periodically." % (path, e),
+                            RuntimeWarning,
+                        )
+
     return fname, cfg
 
 
@@ -185,6 +213,7 @@ def find_connection_file(filename='kernel-*.json', path=None, profile=None):
     for p in path:
         matches.extend(glob.glob(os.path.join(p, pat)))
     
+    matches = [ os.path.abspath(m) for m in matches ]
     if not matches:
         raise IOError("Could not find %r in %r" % (filename, path))
     elif len(matches) == 1:
@@ -276,6 +305,7 @@ class ConnectionFileMixin(LoggingConfigurable):
     _connection_file_written = Bool(False)
 
     transport = CaselessStrEnum(['tcp', 'ipc'], default_value='tcp', config=True)
+    kernel_name = Unicode()
 
     ip = Unicode(config=True,
         help="""Set the kernel\'s IP address [default localhost].
@@ -310,6 +340,9 @@ class ConnectionFileMixin(LoggingConfigurable):
     control_port = Integer(0, config=True,
             help="set the control (ROUTER) port [default: random]")
 
+    # names of the ports with random assignment
+    _random_port_names = None
+
     @property
     def ports(self):
         return [ getattr(self, name) for name in port_names ]
@@ -349,8 +382,9 @@ class ConnectionFileMixin(LoggingConfigurable):
             control_port=self.control_port,
         )
         if session:
-            # add session
-            info['session'] = self.session
+            # add *clone* of my session,
+            # so that state such as digest_history is not shared.
+            info['session'] = self.session.clone()
         else:
             # add session info
             info.update(dict(
@@ -393,6 +427,37 @@ class ConnectionFileMixin(LoggingConfigurable):
             except (IOError, OSError):
                 pass
 
+    def _record_random_port_names(self):
+        """Records which of the ports are randomly assigned.
+
+        Records on first invocation, if the transport is tcp.
+        Does nothing on later invocations."""
+
+        if self.transport != 'tcp':
+            return
+        if self._random_port_names is not None:
+            return
+
+        self._random_port_names = []
+        for name in port_names:
+            if getattr(self, name) <= 0:
+                self._random_port_names.append(name)
+
+    def cleanup_random_ports(self):
+        """Forgets randomly assigned port numbers and cleans up the connection file.
+
+        Does nothing if no port numbers have been randomly assigned.
+        In particular, does nothing unless the transport is tcp.
+        """
+
+        if not self._random_port_names:
+            return
+
+        for name in self._random_port_names:
+            setattr(self, name, 0)
+
+        self.cleanup_connection_file()
+
     def write_connection_file(self):
         """Write connection info to JSON dict in self.connection_file."""
         if self._connection_file_written and os.path.exists(self.connection_file):
@@ -407,6 +472,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             kernel_name=self.kernel_name
         )
         # write_connection_file also sets default ports:
+        self._record_random_port_names()
         for name in port_names:
             setattr(self, name, cfg[name])
 
@@ -443,6 +509,7 @@ class ConnectionFileMixin(LoggingConfigurable):
         self.transport = info.get('transport', self.transport)
         self.ip = info.get('ip', self._ip_default())
 
+        self._record_random_port_names()
         for name in port_names:
             if getattr(self, name) == 0 and name in info:
                 # not overridden by config or cl_args
